@@ -494,14 +494,15 @@ class IrcBot(object):
 
     async def list_names(self, channel):
         """Lists users nicks in channel. 
-        Also check bot.channel_names for a non sanitized version(like starting with @ for operators, if you want to detect them)
+        Also check bot.channel_names for a non sanitized version(like starting with ~ & % @ + for operators, moderators, etc; if you want to detect them)
 
         :param channel:str channel name
         """
         await self._send_data(f"NAMES {channel}")
         await self.sleep(1)
+        special_symbols = ["~", "&", "%", "@", "+",]
         names = self.channel_names[channel]
-        names = [name[1:] if name.startswith("@") else name for name in names]
+        names = [name[1:] if any([name.startswith(s) for s in special_symbols]) else name for name in names]
         return names
 
     async def send_message(self, message, channel=None):
@@ -513,7 +514,6 @@ class IrcBot(object):
         if channel is None:
             channel = self.channels
 
-        log(">>>>>>>>>>>>>>", message)
         if type(channel) == list and type(message) != Message:
             for chan in channel:
                 await self._send_message(message, chan)
@@ -586,7 +586,7 @@ class IrcBot(object):
                 async for data in s:
                     data = data.decode("utf-8")
                     debug(
-                        "DECODED DATA FROM SERVER: \n", 40 * "-", "\n", data, 40 * "-", "\n"
+                        "\nDECODED DATA FROM SERVER: \n", 60 * "-", "\n", data, 60 * "-", "\n"
                     )
                     self.fetch_tables()
                     for msg in data.split("\r\n"):
